package MMM;

use strict;
use warnings;
use MMM::MirrorQueue;
use MMM::MirrorList;
use MMM::Rsync;
use Config::IniFiles;
use POSIX qw(:sys_wait_h);
use File::Temp;
use IO::Select;
use LWP::UserAgent ();
use HTTP::Request ();

our $VERSION = '0.02';

sub new {
    my ($class, %options) = @_;
    my $mmm = {
        configfile => $options{configfile} || 'mmm.cfg',
        mirror_dir => '.',
        loghandle  => $options{loghandle},
        verbosity =>
          ( defined( $options{verbosity} ) ? $options{verbosity} : 3 ),
    };
    $mmm->{config} = Config::IniFiles->new(
        -file => $mmm->{configfile},
        -default => 'default',
    ) or return;
    $mmm->{statedir} = $mmm->{config}->val( 'default', 'statedir', '.' );
    bless( $mmm, $class );
}

sub set_log_handle {
    my ($self, $handle) = @_;
    $self->{loghandle} = $handle;
}

my $loglevel = {
    'PIPE'    => [ 0, '' ],
    'FATAL'   => [ 0, 'Fatal: ' ],
    'ERROR'   => [ 1, 'Error: ' ],
    'WARNING' => [ 2, 'Warning: ' ],
    'NOTICE'  => [ 3, '' ],
    'INFO'    => [ 4, '' ],
    'DEBUG'   => [ 5, 'Debug: ' ],
};

sub log {
    my ( $self, $level, $message, @args ) = @_;
    $message or return;
    my ($keylogl) =
      $level =~ /^\d$/
      ? ( grep { $loglevel->{$_}[0] == $level } keys %{$loglevel} )
      : ( uc($level) );
    $loglevel->{$keylogl} or $keylogl = 'NOTICE';
    $loglevel->{$keylogl}[0] > $self->{verbosity} and return;
    my $h = $self->{loghandle} || \*STDOUT;
    printf $h "$loglevel->{$keylogl}[1]$message\n", @args;
}

sub populate_queue_mirrors {
    my ( $self, %sources ) = @_;
    keys %sources or return 1;
    my $mirror_dir = $self->{config}->val( 'default', 'mirrordir' )
      || $self->{mirror_dir};
    foreach my $xmlfile ( glob("$mirror_dir/*.xml") ) {
        my $ml = MMM::MirrorList->new();
        $ml->load_list($xmlfile);

        foreach my $s ( keys %sources ) {
            foreach my $mirror ( $ml->mirrors($s) ) {
                $sources{$s}->add_mirror($mirror);
            }
        }
    }
    if ( $self->{config}->val( 'default', 'remote_mirrorlist' ) ) {
        my $ua = LWP::UserAgent->new;
        my $http =
          HTTP::Request->new(
            POST => $self->{config}->val( 'default', 'remote_mirrorlist' ) );
        $http->content_type('application/x-www-form-urlencoded');
        $http->content( join( '&', map { "source=$_" } keys %sources ) );

        my $ml = MMM::MirrorList->new();
        if ( my $res = $ua->request($http) ) {
            $ml->load_list( $res->content );
            if ( open( my $mlh, ">", "$self->{statedir}/remotelist.xml" ) ) {
                print $mlh $res->content;
                close($mlh);
            }
        }
        else {
            $ml->load_list("$self->{statedir}/remotelist.xml");
        }

        foreach my $s ( keys %sources ) {
            foreach my $mirror ( $ml->mirrors($s) ) {
                $sources{$s}->add_mirror($mirror);
            }
        }
    }

    1;
}

sub load_mirror_list {
    my ( $self, @queues ) = @_;
    my %sources;
    foreach my $q ( @queues ? @queues : $self->queues ) {
        $sources{ $q->source() } = MMM::MirrorQueue->new( $q->source(), $self );
    }
    $self->populate_queue_mirrors(%sources);
    foreach my $q ( @queues ? @queues : $self->queues ) {
        $q->set_mirror_list( $sources{ $q->source() } );
    }
    1;
}

sub load_one_queue {
    my ($self, $section) = @_;
    my $config = $self->{config} or return;
    my $source = $config->val($section, 'source') or return;
    return MMM::Rsync->new(
        $source,
        dest     => $section,
        statedir => $config->val( 'default', 'statedir', '.' ),
        config   => $self->{config},
        'log' => sub { $self->log(@_) },
    )
}


sub load_queue {
    my ($self) = @_;
    my $config = $self->{config} or return;
    foreach my $dest ($config->Sections()) {
        $dest eq 'default' and next;
        my $mr = $self->load_one_queue($dest);
        $mr or next;
        push (@{$self->{queue}}, $mr);
    }
}

sub load {
    my ($self) = @_;
    $self->load_queue();
    $self->load_mirror_list();
}

sub queues {
    return @{$_[0]->{queue} || []};
}

sub hostname {
    my ($self) = @_;
    $self->{config}->val('default', 'hostname');
}

sub run_one {
    my ($self, $section) = @_;
    my $mr = $self->load_one_queue($section);
    $mr or return;
    $self->load_mirror_list($mr);
    $mr->run();
    if (lc($self->{config}->val('default', 'use_at', 'no')) =~ /(yes|true)/) {
        $self->queue_with_at($mr);
    }
}

sub run {
    my ($self) = @_;
    my $idx = 0;
    my @queue_list;
    foreach (sort {
            my $cmp = $a->average_duration() <=> $b->average_duration();
            $cmp ? $cmp : $a->last_start() <=> $b->last_start()
            } @{$self->{queue} || []}) {
        push(@{$queue_list[$idx]}, $_);
        $idx++;
        $idx=0 if ($idx >= ($self->{config}->val('default', 'parallel') || 2));
    }

    my $ios = IO::Select->new();

    foreach my $q (@queue_list) {
        pipe(my $reader, my $writer);
        my $pid = fork();
        defined($pid) or die("Can't fork");
        if ($pid) {
            $ios->add($reader);
            next;
        } else {
            $self->set_log_handle($writer);
            foreach (@{$q || []}) {
                $_->need_run() or next;
                $_->run();
                if (lc($self->{config}->val('default', 'use_at', 'no')) =~ /(yes|true)/) {
                    $self->queue_with_at($_);
                }
            }
            exit(0);
        }
    }

    my @hs;
    while (@hs = $ios->can_read()) {
        foreach my $h (@hs) {
            my $l = <$h>;
            if(!defined($l)) {
                $ios->remove($h);
                next;
            }
            chomp($l);
            $self->log( 'PIPE', $l );
        }
    }

    my $kid;
    do {
        $kid = waitpid(-1,&WNOHANG);
    } until $kid == -1;

    if ($self->{config}->val('default', 'publish_list')) {
        $self->write_list();
    }
}

sub queue_with_at {
    my ($self, $queue) = @_;
    my $jobf = File::Temp->new(UNLINK => 0);
    if ($jobf) {
        $jobf->unlink_on_destroy( 1 );
        print  $jobf "#!/bin/sh\n";
        printf $jobf "%s run '%s' '%s'\n", $0, $self->{configfile}, $queue->dest;
        close($jobf)
    }
    $ENV{SHELL} = '/bin/sh';
    if (open(my $at,
            sprintf(
                "at -f '%s' + %d minute 2>&1|",
                $jobf->filename(),
                POSIX::ceil(($queue->next_run_time() - scalar(time)) / 60) || 1
            ))) {
        while (<$at>) { }
        close($at);
    }
    print $jobf->filename();
}

sub build_list {
    my ($self) = @_;

    my $ml = MMM::MirrorList->new();

    if ($self->{config}->val('default', 'add_to_publish_list')) {
        $ml->load_list($self->{config}->val('default', 'add_to_publish_list'));
    }

    foreach my $q (@{$self->{queue} || []}) {
        $self->{config}->val($q->dest, 'announce') or next;
        $q->val('subdir') and next; # Don't announce partial tree
        $ml->add_mirror(
            {
                url => $self->{config}->val($q->dest, 'announce'),
                frequency => $q->estimate_frequency || undef,
                level => $q->level,
            },
            $self->{config}->val($q->dest, 'source')
        );
    }

    print $ml->xml_output;
}

sub write_list {
    my ($self, $dest) = @_;
    my $h;
    if (!$dest) {
        if ($self->{config}->val('default', 'publish_list')) {
            open($h, '>', $self->{config}->val('default', 'publish_list')) or return;
        } else {
            return;
        }
    } elsif (ref($dest) eq 'GLOB') {
        $h = $dest;
    } else {
        open($h, '>', $dest) or return;
    }

    print $h $self->build_list();

    if (ref($dest) ne 'GLOB') {
        close($h);
    }
}

1;
