package MMM::MirrorList;

use strict;
use warnings;
use XML::Simple;
use MMM::Mirror;

=head1 NAME

MMM::MirrorList

=head1 SYNOPSIS

    use MMM::MirrorList;
    my $ml = MMM::MirrorList->new();
    $ml->add_mirror({
        url => 'rsync://foo/bar'
        }, 'source');
    print join(' ', $ml->list_sources());

=head1 DESCRIPTION

An object to retain, sort and dump mirrors list.

=head1 METHODS

=head2 new

Return a new MMM::MirrorList object.

=cut

sub new {
    my ($class) = @_;
    bless( {}, $class );
}

=head2 list_sources

Return the source mirror list availlable in this mirrors list.

=cut

sub list_sources {
    my ($self) = @_;
    my %sources;
    foreach ( @{ $self->{list} || [] } ) {
        $sources{ $_->source } = 1;
    }
    return keys %sources;
}

=head2 list_levels

Return the mirrors level availlable in the mirrors list.

=cut

sub list_levels {
    my ($self) = @_;
    my %levels;
    foreach ( @{ $self->{list} || [] } ) {
        $levels{ $_->level } = 1;
    }
    return sort { $a <=> $b } keys %levels;
}

=head2 mirrors($source)

Return a list of L<MMM::Mirror> object handle by this list. If $source
is specified, only mirrors for this source ar returned.

=cut

sub mirrors {
    my ( $self, $source ) = @_;
    return $source
      ? ( grep { $_->source eq $source } @{ $self->{list} || [] } )
      : ( @{ $self->{list} || [] } );
}

=head2 add_mirror($info, $souce)

Add a new mirror into the this mirror list.

$info can either a L<MMM::Mirror> object or a hash containing mirror
information.

$source is the source name.

=cut

sub add_mirror {
    my ( $self, $infos, $source ) = @_;
    $source ||= '-';
    my $mirror =
      ref( $infos || {} ) eq 'HASH' ? MMM::Mirror->new( %{$infos} ) : $infos;
    $mirror or return;
    if ($source) {
        $mirror->set_source($source);
    }
    else {
        $source = $mirror->source || '-';
    }

    # avoid twice mirror in list
    foreach my $m ( $self->mirrors() ) {
        if ( $m->same_mirror($mirror) ) {
            $m->sync_mirror($mirror);
            return 1;
        }
    }
    push( @{ $self->{list} }, $mirror );
    return 1;
}

=head2 load_list($xml)

Load mirrors from the xml mirrors list. $xml is either the xml string
or the xml file.

=cut

sub load_list {
    my ( $self, $xmlfile ) = @_;
    my $ref = XMLin(
        $xmlfile,

        #KeyAttr => { 'source' => '+name' },
        ForceArray => [ 'mirror', 'source' ],
    ) or return;

    foreach my $source ( keys %{ $ref->{source} || {} } ) {
        foreach my $mirror ( @{ $ref->{source}{$source}{mirror} || [] } ) {
            $self->add_mirror( $mirror, $source );
        }
    }

    1;
}

=head2 xml_output

Return an xml string about mirror in mirrors list.

=cut

sub xml_output {
    my ($self) = @_;

    my $xml = '<?xml version="1.0" encoding="UTF-8"?>';
    $xml .= "\n<mmm>\n";
    foreach my $source ( $self->list_sources ) {
        $xml .= "\t" . sprintf( '<source name="%s">', $source ) . "\n";
        foreach ( $self->mirrors($source) ) {
            $xml .= $_->xml_output();
        }
        $xml .= "\t</source>\n";
    }
    $xml .= "</mmm>\n";

    $xml;
}

1;

=head1 SEE ALSO

L<MMM>
L<MMM::Mirror>

=head1 AUTHOR

Olivier Thauvin <nanardon@nanardon.zarb.org>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2006 Olivier Thauvin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

=cut

